#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define MAXBUFFER	512
#define BPARRAFO	16
#define MAXSEGMENTO	65535
#define MARCAEJEC   0x5643

typedef struct {
	unsigned marca;		/* Marca de ejecutable */
	unsigned tamcodigo;	/* Bytes del segmento de codigo */
	unsigned tamdatos;	/* Bytes de datos inicializados */
	unsigned tambss;		/* Bytes de bss */
	unsigned tamtotal;	/* Bytes totales asignados al programa */
	char reservado[22];
} CabPrograma;

long ConvHexToDec(const char *buff, int index, int longitud)
{
	long result = 0;
	int i, digito;

	for (i = index; i < index + longitud; i++) {
		if(isdigit(buff[i])) digito = buff[i] - '0';
		else {
			digito = tolower(buff[i]) - 'a' + 10;
			if(digito > 15 || digito < 10) return -1;
		}
		result = 16 * result + digito;
	}
	return result;
}

int ObtenerTamannos(const char *arcmap, long *ptamcodigo, long *ptamdatos, long *ptambss)
{
	FILE *pf;
	char buffer[256];

	if((pf = fopen(arcmap, "r")) == NULL) return 1;


	for( ; ; ) { /* Obtener la longitud del segmento de texto */
		if(fgets(buffer, 256, pf) == NULL) {
			fclose(pf);
			return 2;
		}
		if (strstr(buffer, "_TEXT") != NULL) {
			if((*ptamcodigo = ConvHexToDec(buffer, 15, 5)) < 0) {
				fclose(pf);
				return 2;
			}
			break;
		}
	}


	rewind(pf);
	for( ; ; ) { /* Obtener la longitud del segmento de datos */
		if(fgets(buffer, 256, pf) == NULL) {
			fclose(pf);
			return 3;
		}
		if (strstr(buffer, "_DATA") != NULL) {
			if((*ptamdatos = ConvHexToDec(buffer, 15, 5)) < 0) {
				fclose(pf);
				return 3;
			}
			break;
		}
	}


	rewind(pf);
	for( ; ; ) { /* Obtener la longitud del segmento de BSS */
		if(fgets(buffer, 256, pf) == NULL) {
			fclose(pf);
			return 4;
		}
		if (strstr(buffer, "_BSS") != NULL) {
			if((*ptambss = ConvHexToDec(buffer, 15, 5)) < 0) {
				fclose(pf);
				return 4;
			}
			break;
		}
	}
	return 0;
}

int main(int argc, char *argv[])
{
	long tamcodigo, tamdatos, tambss, tamtotal;
	FILE *forigen, *fdestino;
	CabPrograma cab;
	char buffer[MAXBUFFER];
	int num;

	/* Comprobacion numero de argumentos */
	if (argc != 4) {
		fprintf(stderr, "Uso: conv <arc_origen> <arc_map> <arc_destino>\n");
		return 1;
	}

	/* Comprobacion de archivos */
	strlwr(argv[1]); strlwr(argv[3]);
	if (strcmp(argv[1], argv[3]) == 0) {
		fprintf(stderr, "Error: archivo destino igual a archivo origen\n");
		return 1;
	}

	/* Obtener tamannos de memoria del archivo map */
	if (ObtenerTamannos(argv[2], &tamcodigo, &tamdatos, &tambss))
		fprintf(stderr, "Error: no se ha podido obtener la informacin del archivo %s", argv[2]);

	/* Comprobacion de los valores de la memoria solicitada */
	tamtotal = tamcodigo + tamdatos + tambss;
	if ((tamtotal % BPARRAFO) != 0)
		tamtotal = (tamtotal/BPARRAFO)*BPARRAFO + BPARRAFO;
	if ((tamcodigo <= 0) || (tamcodigo > MAXSEGMENTO)) {
		fprintf(stderr, "Memoria para codigo incorrecta\n");
		return 1;
	}
	if ((tamdatos < 0) || (tamdatos > MAXSEGMENTO)) {
		fprintf(stderr, "Memoria para datos incorrecta\n");
		return 1;
	}
	if ((tambss < 0) || (tambss > MAXSEGMENTO)) {
		fprintf(stderr, "Memoria para bbs incorrecta\n");
		return 1;
	}
	if (tamtotal > MAXSEGMENTO) {
		fprintf(stderr, "Memoria total excede un segmento\n");
		return 1;
	}

	/* Apertura de archivos */
	if ((forigen = fopen(argv[1], "rb")) == NULL) {
		fprintf(stderr, "Error en archivo origen\n");
		return 1;
	}
	if ((fdestino = fopen(argv[3], "wb")) == NULL) {
		fclose(forigen);
		fprintf(stderr, "Error en archivo destino\n");
		return 1;
	}

	cab.marca = MARCAEJEC;
	cab.tamcodigo = tamcodigo;
	cab.tamdatos = tamdatos;
	cab.tambss = tambss;
	cab.tamtotal = tamtotal;
	if (fwrite(&cab, sizeof(cab), 1, fdestino) != 1) {
		fprintf(stderr, "Error en archivo destino\n");
		fclose(forigen); fclose(fdestino);
		return 1;
	}
	while (feof(forigen) == 0) {
		num = fread(buffer, 1, MAXBUFFER, forigen);
		fwrite(buffer, 1, num, fdestino);
	}
	fclose(forigen);
	fclose(fdestino);
	return 0;
}
